# document.py

# noinspection PyUnresolvedReferences
from multithreading.process import *

from multithreading.base import document

class Documentation:
    """"""
# end Documentation

document(Documentation)